from ingenialink.utils._utils import raise_err
from ingenialink.poller import Poller
from .constants import *

from datetime import datetime
from threading import Timer, RLock

import ingenialogger
logger = ingenialogger.get_logger(__name__)


class PollerTimer:
    """Custom timer for the CanopenPoller.

    Args:
        time (int): Timeout to use for the timer.
        cb (function): Callback.

    """
    def __init__(self, time, cb):
        self.cb = cb
        self.time = time
        self.thread = Timer(self.time, self.handle_function)

    def handle_function(self):
        """Handle method that creates the timer for the poller"""
        self.cb()
        self.thread = Timer(self.time, self.handle_function)
        self.thread.start()

    def start(self):
        """Starts the poller timer"""
        self.thread.start()

    def cancel(self):
        """Stops the poller timer"""
        self.thread.cancel()
        if self.thread.is_alive():
            self.thread.join()


class CanopenPoller(Poller):
    """Register poller for CANOpen communications.

    Args:
        servo (CanopenServo): Servo.
        num_channels (int): Number of channels.

    Raises:
        ILCreationError: If the poller could not be created.

    """

    def __init__(self, servo, num_channels):
        super(CanopenPoller, self).__init__(servo, num_channels)
        self.__sz = 0
        self.__refresh_time = 0
        self.__time_start = 0.0
        self.__samples_count = 0
        self.__samples_lost = False
        self.__timer = None
        self.__running = False
        self.__mappings = []
        self.__mappings_enabled = []
        self.__lock = RLock()
        self._reset_acq()

    def start(self):
        """Start the poller."""

        if self.__running:
            logger.warning("Poller already running")
            raise_err(IL_EALREADY)

        # Activate timer
        self.__timer = PollerTimer(self.__refresh_time,
                                   self._acquire_callback_poller_data)
        self.__timer.start()
        self.__time_start = datetime.now()

        self.__running = True

        return 0

    def stop(self):
        """Stop poller."""

        if self.__running:
            self.__timer.cancel()

        self.__running = False

    def configure(self, t_s, sz):
        """Configure data.

        Args:
            t_s (int, float): Polling period (s).
            sz (int): Buffer size.

        Returns:
            int: Status code.

        Raises:
            ILStateError: The poller is already running.

        """
        if self.__running:
            logger.warning("Poller is running")
            raise_err(IL_ESTATE)

        # Configure data and sizes with empty data
        self._reset_acq()
        self.__sz = sz
        self.__refresh_time = t_s
        self.__acq['t'] = [0] * sz
        for channel in range(0, self.num_channels):
            data_channel = [0] * sz
            self.__acq['d'].append(data_channel)
            self.__mappings.append('')
            self.__mappings_enabled.append(False)

        return 0

    def ch_configure(self, channel, reg):
        """Configure a poller channel mapping.

        Args:
            channel (int): Channel to be configured.
            reg (Register): Register to associate to the given channel.

        Returns:
            int: Status code.

        Raises:
            ILStateError: The poller is already running.
            ILValueError: Channel out of range.
            TypeError: If the register is not valid.

        """

        if self.__running:
            logger.warning("Poller is running")
            raise_err(IL_ESTATE)

        if channel > self.num_channels:
            logger.error("Channel out of range")
            raise_err(IL_EINVAL)

        # Obtain register
        _reg = self.servo._get_reg(reg)

        # Reg identifier obtained and set enabled
        self.__mappings[channel] = {}
        self.__mappings[channel][_reg.identifier] = int(_reg.subnode)
        self.__mappings_enabled[channel] = True

        return 0

    def ch_disable(self, channel):
        """Disable a channel.

        Args:
            channel (int): Channel to be disabled.

        Raises:
            ILStateError: The poller is already running.
            ILValueError: Channel out of range.

        Returns:
            int: Status code.

        """

        if self.__running:
            logger.warning("Poller is running")
            raise_err(IL_ESTATE)

        if channel > self.num_channels:
            logger.error("Channel out of range")
            raise_err(IL_EINVAL)

        # Set channel required as disabled
        self.__mappings_enabled[channel] = False

        return 0

    def ch_disable_all(self):
        """Disable all channels.

        Returns:
            int: Status code.

        """

        for channel in range(0, self.num_channels):
            r = self.ch_disable(channel)
            if r < 0:
                raise_err(r)
        return 0
    
    def _reset_acq(self):
        """Resets the aquired channels."""
        self.__acq = {
            "t": [],
            "d": []
        }

    def _acquire_callback_poller_data(self):
        """Acquire callback for poller data."""
        time_diff = datetime.now()
        delta = time_diff - self.__time_start

        # Obtain current time
        t = delta.total_seconds()

        self.__lock.acquire()
        # Acquire all configured channels
        if self.__samples_count >= self.__sz:
            self.__samples_lost = True
        else:
            self.__acq['t'][self.__samples_count] = t

            # Acquire enabled channels, comprehension list indexes obtained
            enabled_channel_indexes = [
                channel_idx for channel_idx, is_enabled in
                enumerate(self.__mappings_enabled) if is_enabled
            ]

            for channel in enabled_channel_indexes:
                for register_identifier, subnode in \
                        self.__mappings[channel].items():
                    self.__acq['d'][channel][self.__samples_count] = \
                        self.servo.read(register_identifier, subnode)

            # Increment samples count
            self.__samples_count += 1

        self.__lock.release()
    
    @property
    def data(self):
        """tuple (list, list, bool): Time vector, array of data vectors and a
            flag indicating if data was lost."""
        t = list(self.__acq['t'][0:self.__samples_count])
        d = []

        # Acquire enabled channels, comprehension list indexes obtained
        enabled_channel_indexes = [
            channel_idx for channel_idx, is_enabled in
            enumerate(self.__mappings_enabled) if is_enabled
        ]

        for channel in range(0, self.num_channels):
            if self.__mappings_enabled[channel]:
                d.append(
                    list(self.__acq['d'][channel][0:self.__samples_count])
                )
            else:
                d.append(list(None))

        self.__lock.acquire()
        self.__samples_count = 0
        self.__samples_lost = False
        self.__lock.release()

        return t, d, self.__samples_lost



#!/usr/bin/env python

import re
from setuptools import setup
from distutils.cmd import Command
import shutil
import os

_version = re.search(r'__version__\s+=\s+\'(.*)\'',
                     open('ingenialink/__init__.py').read()).group(1)


def get_docs_url():
    return "https://distext.ingeniamc.com/doc/ingenialink-python/{}".format(_version)


class BDistAppCommand(Command):
    """Custom command to build the application."""

    description = 'Build the application'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Copying dlls...")
        shutil.copy('resources/Packet.dll', 'ingenialink/')
        shutil.copy('resources/wpcap.dll', 'ingenialink/')


class BCleanAppCommand(Command):
    """Custom command to clean the application."""

    description = 'Clean the application'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("Cleaning files...")
        if os.path.exists("ingenialink/Packet.dll"):
            os.remove("ingenialink/Packet.dll")
        if os.path.exists("ingenialink/wpcap.dll"):
            os.remove("ingenialink/wpcap.dll")


setup(name='ingenialink',
      version=_version,
      packages=['ingenialink', 'ingenialink.canopen', 'ingenialink.ethercat',
                'ingenialink.ethernet', 'ingenialink.serial', 'ingenialink.ipb',
                'ingenialink.utils'],
      description='IngeniaLink Communications Library',
      long_description=open('README.rst').read(),
      author='Ingenia Motion Control',
      author_email='support@ingeniamc.com',
      url='https://www.ingeniamc.com',
      project_urls={
          'Documentation': get_docs_url(),
          'Source': 'https://github.com/ingeniamc/ingenialink-python'
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS',
          'Operating System :: Microsoft :: Windows',
          'Programming Language :: C',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Communications',
          'Topic :: Software Development :: Libraries'
      ],
      cmdclass={
        'bdist_app': BDistAppCommand,
        'bclean_app': BCleanAppCommand
      },
      setup_requires=['cffi==1.14.6'],
      cffi_modules=['ingenialink/ingenialink_build.py:ffibuilder'],
      install_requires=[
          'cffi==1.14.6',
          'numpy<=1.19.5',
          'canopen>=1.0.0',
          'ingenialogger>=0.2.1'
      ],
      include_package_data = True

      )



from .._ingenialink import lib, ffi
from ingenialink.network import NET_PROT
from ingenialink.exceptions import ILError
from ingenialink.ipb.network import IPBNetwork
from ingenialink.serial.servo import SerialServo
from ingenialink.utils._utils import pstr, cstr, raise_null, to_ms


class SerialNetwork(IPBNetwork):
    """Network for all Serial communications.

    Args:
        timeout_rd (int): Timeout for read commands.
        timeout_wr (int): Timeout for write commands.

    """
    def __init__(self, timeout_rd=0.5, timeout_wr=0.5):
        super(SerialNetwork, self).__init__()
        self.timeout_rd = timeout_rd
        self.timeout_wr = timeout_wr

        self.__net_interface = None

    def load_firmware(self, fw_file):
        raise NotImplementedError

    def scan_slaves(self):
        """Obtain a list of network devices.

        Returns:
            list: List of network devices.

        Raises:
            TypeError: If the protocol type is invalid.
        """
        devs = lib.il_net_dev_list_get(NET_PROT.MCB.value)

        found = []
        curr = devs

        while curr:
            found.append(pstr(curr.port))
            curr = curr.next

        lib.il_net_dev_list_destroy(devs)

        return found

    def connect_to_slave(self, target, dictionary=""):
        """Connects to a slave through the given network settings.

        Args:
            target (str): IP of the target slave.
            dictionary (str): Path to the target dictionary file.

        Returns:
            SerialServo: Instance of the servo connected.

        """
        self._on_found = ffi.NULL

        callback = ffi.NULL
        handle = ffi.NULL

        opts = ffi.new('il_net_opts_t *')
        _port = ffi.new('char []', cstr(target))
        opts.port = _port
        opts.timeout_rd = to_ms(self.timeout_rd)
        opts.timeout_wr = to_ms(self.timeout_wr)

        self.__net_interface = lib.il_net_create(NET_PROT.MCB.value, opts)
        raise_null(self.__net_interface)

        servos = lib.il_net_servos_list_get(self.__net_interface, callback, handle)

        found = []
        curr = servos

        while curr:
            found.append(int(curr.id))
            curr = curr.next

        lib.il_net_servos_list_destroy(servos)

        if not found:
            raise ILError('Could not connect to "{}"'.format(target))

        servo = SerialServo(self.__net_interface, target, found[0], dictionary)
        self._cffi_network = self.__net_interface
        self.servos.append(servo)
        return servo

    def disconnect_from_slave(self, servo):
        """Disconnects the slave from the network.

        Args:
            servo (SerialServo): Instance of the servo connected.

        """
        self.servos.remove(servo)
        if len(self.servos) == 0:
            lib.il_net_disconnect(self._cffi_network)
        self._cffi_network = None

    @property
    def protocol(self):
        """NET_PROT: Obtain network protocol."""
        return NET_PROT.MCB
